from bovine_store.utils.test import (  # noqa F401
    store,
    bovine_store_actor,
    bovine_admin_store,
)

from bovine_process.types import ProcessingItem

from .interactions import reply_handler


async def test_reply_handle_bare_object(store, bovine_store_actor):  # noqa F801
    remote_actor = "https://remote_domain/actor"
    first_id = "https://remote_domain/first"
    second_id = "https://remote_domain/second"
    create = {
        "@context": "about:bovine",
        "id": first_id,
        "type": "Create",
        "actor": remote_actor,
        "object": second_id,
    }

    processing_item = ProcessingItem(remote_actor, create)

    await reply_handler(processing_item, bovine_store_actor)



import logging

from bovine_store.utils import add_sub_collections

from bovine_process.types import ProcessingItem

logger = logging.getLogger(__name__)


async def update_id(item: ProcessingItem, actor) -> ProcessingItem:
    """Assigns a new id to the object in data and possibly to the
    object in data["object"]"""
    data = await update_id_function(item.data, actor)
    item.data = data

    item.meta["object_location"] = data["id"]

    return item


async def update_id_function(data: dict, actor):
    data["id"] = actor.generate_new_object_id()
    if "object" in data and isinstance(data["object"], dict):
        if "id" in data["object"]:
            if data["object"]["id"] == actor.actor_id:
                return data

            obj_in_store = await actor.retrieve(data["object"]["id"])
            if not obj_in_store:
                data["object"]["id"] = actor.generate_new_object_id()
        else:
            data["object"]["id"] = actor.generate_new_object_id()

        data["object"] = add_sub_collections(data["object"])

    return data



import logging

from bovine.activitystreams.utils import actor_for_object

from bovine_process.types import ProcessingItem

logger = logging.getLogger(__name__)


async def handle_update(item: ProcessingItem, actor) -> ProcessingItem:
    owner = actor_for_object(item.data)

    object_to_update = item.data.get("object")
    if object_to_update is None or object_to_update.get("id") is None:
        logger.warning("Update without object %s", item.body)
        return

    to_update_from_db = await actor.retrieve_for(owner, object_to_update["id"])
    object_to_update["@context"] = item.data["@context"]

    if to_update_from_db:
        await actor.update_for(owner, object_to_update)

    return item



from bovine_process.utils.processor_list import ProcessorList
from bovine.jsonld import with_bovine_context

from .following import accept_follow, undo_follow
from .outgoing_delete import outgoing_delete
from .outgoing_update import outgoing_update
from .store_outgoing import add_outgoing_to_outbox, store_outgoing
from .update_id import update_id


async def sanitize(item, actor):
    item.data = with_bovine_context(item.data)
    return item


default_outbox_process = (
    ProcessorList()
    .add(sanitize)
    .add(update_id)
    .add(store_outgoing)
    .add(add_outgoing_to_outbox)
    .add_for_types(
        Update=outgoing_update,
        Delete=outgoing_delete,
    )
    .apply
)
"""Defines the synchrnous part of sending an outgoing object"""


default_async_outbox_process = (
    ProcessorList().add_for_types(Accept=accept_follow, Undo=undo_follow).apply
)



from bovine_store.utils.test import (  # noqa F401
    store,
    bovine_store_actor,
    bovine_admin_store,
)


from bovine_process.types import ProcessingItem

from .outgoing_update import outgoing_update
from .store_outgoing import store_outgoing


async def test_basic_update(store, bovine_store_actor):  # noqa F801
    first_id = "https://my_domain/first"
    second_id = "https://my_domain/second"
    third_id = "https://my_domain/third"
    create = {
        "@context": "https://www.w3.org/ns/activitystreams",
        "id": first_id,
        "type": "Create",
        "actor": bovine_store_actor.actor_id,
        "object": {
            "type": "Note",
            "id": second_id,
            "content": "new",
            "attributedTo": bovine_store_actor.actor_id,
        },
    }

    update = {
        "@context": "https://www.w3.org/ns/activitystreams",
        "id": third_id,
        "type": "Create",
        "actor": bovine_store_actor.actor_id,
        "object": {
            "type": "Note",
            "id": second_id,
            "content": "updated",
            "attributedTo": bovine_store_actor.actor_id,
        },
    }

    processing_item = ProcessingItem(bovine_store_actor.actor_id, create)

    await store_outgoing(processing_item, bovine_store_actor)

    stored = await store.retrieve(bovine_store_actor.actor_id, second_id)
    assert stored["content"] == "new"

    processing_item = ProcessingItem(bovine_store_actor.actor_id, update)
    await outgoing_update(processing_item, bovine_store_actor)

    stored = await store.retrieve(bovine_store_actor.actor_id, second_id)
    assert stored["content"] == "updated"



from bovine_store.utils.test import (  # noqa F401
    store,
    bovine_store_actor,
    bovine_admin_store,
)


from bovine_process.types import ProcessingItem

from .following import accept_follow, undo_follow


async def test_accept_follow(store, bovine_store_actor):  # noqa F801
    remote_actor = "https://remote/actor"
    actor_object = bovine_store_actor.actor_object

    activity = {
        "@context": "about:bovine",
        "type": "Accept",
        "actor": remote_actor,
        "object": {
            "id": actor_object.id + "/follow",
            "actor": actor_object.id,
            "type": "Follow",
            "object": remote_actor,
        },
    }
    item = ProcessingItem(remote_actor, activity)

    await accept_follow(item, bovine_store_actor)

    following = await bovine_store_actor.resolve_endpoints({actor_object.following})

    assert remote_actor in following


async def test_undo_follow(store, bovine_store_actor):  # noqa F801
    remote_actor = "https://remote/actor"
    actor_object = bovine_store_actor.actor_object

    await bovine_store_actor.add_to_followers(remote_actor)

    activity = {
        "@context": "about:bovine",
        "type": "Undo",
        "actor": remote_actor,
        "object": {"actor": remote_actor, "type": "Follow", "object": actor_object.id},
    }
    item = ProcessingItem(remote_actor, activity)

    await undo_follow(item, bovine_store_actor)

    followers = await bovine_store_actor.resolve_endpoints({actor_object.followers})

    assert remote_actor not in followers



from unittest.mock import AsyncMock

from bovine_process.types import ProcessingItem

from .by_activity_type import ByActivityType, do_nothing


async def test_do_nothing():
    item = ProcessingItem("submitter", "item")

    assert await do_nothing(item, "other arg") == item
    assert await do_nothing(item, "other arg", {"more": "arguments"}) == item


async def test_by_activity_type():
    item = ProcessingItem("submitter", {"type": "Test"})

    mock = AsyncMock()
    mock.return_value = "mock"

    by_activity_type = ByActivityType({"Test": mock})

    result = await by_activity_type.act(item)

    assert result == "mock"
    mock.assert_awaited_once()


async def test_build_do_for_types():
    follow_item = ProcessingItem("submitter", {"type": "Follow"})
    create_item = ProcessingItem("submitter", {"type": "Create"})

    mock = AsyncMock()
    mock.return_value = "mock"

    processor = ByActivityType({"Follow": mock}).act

    assert await processor(follow_item) == "mock"
    assert await processor(create_item) == create_item

    mock.assert_awaited_once()



import logging

from bovine.jsonld import with_bovine_context

from bovine_process.utils.processor_list import ProcessorList

from .following import accept_follow, undo_follow
from .handle_update import handle_update
from .incoming_delete import incoming_delete
from .interactions import (
    announce_handler,
    delete_reply_handler,
    like_handler,
    reply_handler,
    undo_handler,
)
from .store_incoming import add_incoming_to_inbox, store_incoming

logger = logging.getLogger(__name__)


async def sanitize(item, actor):
    item.data = with_bovine_context(item.data)

    if item.submitter != item.data["actor"]:
        logger.error("Got wrong submitter for an activity %s", item.submitter)
        logger.error(item.data)
        # return

    return item


interaction_handlers = {
    **dict(
        Announce=announce_handler,
        Create=reply_handler,
        Delete=delete_reply_handler,
        Dislike=like_handler,
        Like=like_handler,
        Undo=undo_handler,
    ),
    "http://litepub.social/ns#EmojiReact": like_handler,
}
"""The handlers being called for interactions"""

crud_handlers = dict(Update=handle_update, Delete=incoming_delete)
"""The handlers being called for CRUD operations"""

social_handlers = dict(Accept=accept_follow, Undo=undo_follow)
"""The handlers being called for social interactions, i.e. updating the social graph

FIXME: Missing undo_follow ... and the other nonsense, I need somebody to draw me
a state machine for"""


default_inbox_process = (
    ProcessorList()
    .add(sanitize)
    .add(store_incoming)
    .add(add_incoming_to_inbox)
    .add_for_types(**crud_handlers)
    .add_for_types(**social_handlers)
    .add_for_types(**interaction_handlers)
    .apply
)
"""Represents the default process undertaken by an inbox item"""



import logging

from bovine_process.types import ProcessingItem

logger = logging.getLogger(__name__)


async def outgoing_update(item: ProcessingItem, actor) -> ProcessingItem:
    object_to_update = item.data.get("object")
    object_to_update["@context"] = item.data["@context"]
    if object_to_update is None or object_to_update.get("id") is None:
        logger.warning("Update without object %s", item.body)
        return

    if object_to_update.get("id") == actor.actor_id:
        logger.info("Updating actor")
        await actor.update_profile(object_to_update)
        item.meta["additional_recipients"] = {actor.actor_object.followers}
        return item

    to_update_from_db = await actor.retrieve(
        object_to_update.get("id"), skip_fetch=True
    )

    if to_update_from_db.get("attributedTo") != actor.actor_id:
        logger.warning(
            "Update from different actor %s in %s for object with id %s",
            actor.actor_id,
            item.data.get("id"),
            object_to_update.get("id"),
        )
        return

    # FIXME: More sanity checks?

    if to_update_from_db:
        await actor.update(object_to_update)

    return item



import uuid
from unittest.mock import AsyncMock, MagicMock

from .update_id import update_id_function


async def test_update_id():
    actor = AsyncMock()
    actor.generate_new_object_id = lambda: str(uuid.uuid4())

    data = {"id": "id"}
    result = await update_id_function(data, actor)
    assert result["id"] != "id"

    data = {"object": "test"}
    result = await update_id_function(data, actor)
    assert result["id"]

    # if in store; id is not updated
    data = {"object": {"id": "id"}}
    result = await update_id_function(data, actor)
    assert result["object"]["id"] == "id"

    actor.retrieve.return_value = None
    data = {"object": {"id": "id"}}
    result = await update_id_function(data, actor)
    assert result["object"]["id"] != "id"


async def test_update_id_for_actor_update():
    actor = MagicMock(actor_id="my_actor")
    actor.generate_new_object_id = lambda: str(uuid.uuid4())

    data = {"type": "Update", "object": {"id": actor.actor_id}}

    data = await update_id_function(data, actor)

    assert data["object"]["id"] == actor.actor_id



import logging
import traceback

logger = logging.getLogger(__name__)


async def do_nothing(item, *args):
    return item


class ByActivityType:
    def __init__(self, actions: dict):
        self.actions = actions

    async def act(self, item, *args):
        try:
            item_type = item.data["type"]
            if item_type in self.actions:
                return await self.actions[item_type](item, *args)
            else:
                return item
        except Exception as ex:
            logger.error(f"Something went wrong with {ex} during procession")
            if item:
                logger.error(item.data)

            for log_line in traceback.format_exc().splitlines():
                logger.error(log_line)



import logging

from bovine.activitystreams.object_factory import Object
from bovine.activitystreams.utils import actor_for_object
from bovine.jsonld import with_bovine_context
from bovine_process.types import ProcessingItem

logger = logging.getLogger(__name__)


async def incoming_delete(item: ProcessingItem, actor) -> ProcessingItem:
    owner = actor_for_object(item.data)
    object_to_delete = item.data.get("object")

    if isinstance(object_to_delete, dict):
        object_to_delete = object_to_delete.get("id")

    if object_to_delete is None:
        logger.warning("Delete without object %s", item.body)
        return

    to_update_from_db = await actor.retrieve_for(owner, object_to_delete)

    tombstone = with_bovine_context(
        Object(type="Tombstone", id=object_to_delete).build()
    )

    if to_update_from_db:
        await actor.update_for(owner, tombstone)

    return item
    # return None



import logging

from bovine.parse import Activity
from bovine_process.types import ProcessingItem

logger = logging.getLogger(__name__)


async def accept_follow(item: ProcessingItem, actor) -> ProcessingItem:
    accept = Activity(item.data, domain=item.submitter_domain)
    follow = await accept.accept_for_follow(actor.retrieve)

    if follow and follow.actor_id == actor.actor_id:
        await actor.add_to_following(follow.object_id)
        logger.info(
            "Added %s to following %s", follow.object_id, actor.actor_object.following
        )

    return item


async def undo_follow(item: ProcessingItem, actor) -> ProcessingItem:
    undo = Activity(item.data, domain=item.submitter_domain)
    follow = await undo.undo_of_follow(actor.retrieve)

    if follow:
        await actor.remove_from_followers(follow.actor_id)
        logger.info(
            "Removed %s from followers %s",
            follow.actor_id,
            actor.actor_object.followers,
        )

    return item



import os
import json
import aiohttp
import logging

from faststream import FastStream, Context
from faststream.rabbit import RabbitBroker, RabbitExchange, RabbitQueue, ExchangeType
from faststream.annotations import Logger, ContextRepo


from bovine_process.types import ProcessingItem
from bovine_process import process_inbox_item
from bovine_process.outgoing import default_async_outbox_process
from bovine_process.send_item import (
    determine_recipients,
    send_to_inbox,
    get_inbox_for_recipient,
)
from bovine.types import ServerSentEvent, Visibility
from bovine.jsonld import with_external_context

from bovine_store.config import tortoise_config
from tortoise import Tortoise

from bovine_propan.actor import bovine_actor

logging.basicConfig(level=logging.INFO)

broker = RabbitBroker(os.environ.get("BOVINE_AMQP", "amqp://localhost"))
app = FastStream(broker)
"""faststream app can be run using 

.. code-block:: bash

    faststream run bovine_propan/__init__:app"""

exch = RabbitExchange("processing", auto_delete=False, type=ExchangeType.TOPIC)
processed = RabbitExchange("processed", auto_delete=False, type=ExchangeType.TOPIC)

inbox_queue = RabbitQueue("inbox", auto_delete=False, routing_key="inbox")
outbox_queue = RabbitQueue("outbox", auto_delete=False, routing_key="outbox")
to_send_queue = RabbitQueue("to_send", auto_delete=False, routing_key="to_send")


@app.on_startup
async def startup(context: ContextRepo):
    db_url = os.environ.get("BOVINE_DB_URL", "sqlite://bovine.sqlite3")
    await Tortoise.init(config=tortoise_config(db_url))
    await Tortoise.generate_schemas()

    session = aiohttp.ClientSession()
    context.set_global("session", session)


@app.on_shutdown
async def shutdown(session: aiohttp.ClientSession = Context()):
    await Tortoise.close_connections()
    session.close()


@broker.subscriber(inbox_queue, exch)
async def inbox_handler(
    body: dict, logger: Logger, session: aiohttp.ClientSession = Context()
):
    async with bovine_actor(body, session) as actor:
        item = ProcessingItem(body["submitter"], body["data"])
        await process_inbox_item(item, actor)

        data_s = json.dumps(item.data)
        event = ServerSentEvent(data=data_s, event="inbox")

        if "database_id" in item.meta:
            event.id = item.meta["database_id"]

        actor_info = actor.actor_object.build(visibility=Visibility.OWNER)
        event_source = actor_info["endpoints"]["eventSource"]

        await broker.publish(
            event.encode(), routing_key=event_source, exchange=processed
        )


@broker.subscriber(outbox_queue, exch)
async def outbox_handler(
    body: dict, logger: Logger, session: aiohttp.ClientSession = Context()
):
    async with bovine_actor(body, session) as actor:
        item = ProcessingItem(body["submitter"], body["data"])
        await default_async_outbox_process(item, actor)

        recipients = await determine_recipients(item, actor)
        to_send = with_external_context(item.data)

        for recipient in recipients:
            await broker.publish(
                {
                    "recipient": recipient,
                    "data": to_send,
                    "bovine_name": body["bovine_name"],
                },
                routing_key="to_send",
                exchange=exch,
            )

        data_s = json.dumps(item.data)
        event = ServerSentEvent(data=data_s, event="inbox")

        if "database_id" in item.meta:
            event.id = item.meta["database_id"]

        actor_info = actor.actor_object.build(visibility=Visibility.OWNER)
        event_source = actor_info["endpoints"]["eventSource"]

        await broker.publish(
            event.encode(), routing_key=event_source, exchange=processed
        )


@broker.subscriber(to_send_queue, exch)
async def to_send_handler(
    body: dict, logger: Logger, session: aiohttp.ClientSession = Context()
):
    async with bovine_actor(body, session) as actor:
        inbox = await get_inbox_for_recipient(actor, body["recipient"])
        if inbox:
            await send_to_inbox(actor, inbox, body["data"])



from bovine_store.utils.test import (  # noqa F401
    store,
    bovine_store_actor,
    bovine_admin_store,
)

from bovine_process.types import ProcessingItem

from .handle_update import handle_update
from .store_incoming import store_incoming


async def test_basic_update(store, bovine_store_actor):  # noqa F801
    remote_actor = "https://remote_domain/actor"
    first_id = "https://my_domain/first"
    second_id = "https://my_domain/second"
    third_id = "https://my_domain/third"
    create = {
        "@context": "https://www.w3.org/ns/activitystreams",
        "id": first_id,
        "type": "Create",
        "actor": remote_actor,
        "object": {"type": "Note", "id": second_id, "content": "new"},
    }

    update = {
        "@context": "https://www.w3.org/ns/activitystreams",
        "id": third_id,
        "type": "Create",
        "actor": remote_actor,
        "object": {"type": "Note", "id": second_id, "content": "updated"},
    }

    actor_id = bovine_store_actor.actor_object.id

    processing_item = ProcessingItem(remote_actor, create)

    await store_incoming(processing_item, bovine_store_actor)

    stored = await store.retrieve(actor_id, second_id)
    assert stored["content"] == "new"

    processing_item = ProcessingItem(remote_actor, update)
    await handle_update(processing_item, bovine_store_actor)

    stored = await store.retrieve(actor_id, second_id)
    assert stored["content"] == "updated"



import logging

from bovine.activitystreams.object_factory import Object
from bovine.activitystreams.utils import copy_to_and_cc

from bovine_process.types import ProcessingItem

logger = logging.getLogger(__name__)


async def outgoing_delete(item: ProcessingItem, actor) -> ProcessingItem:
    object_to_delete = item.data.get("object")

    if isinstance(object_to_delete, dict):
        object_to_delete = object_to_delete.get("id")

    if object_to_delete is None:
        logger.warning("Delete without object %s", item.body)
        return None

    to_update_from_db = await actor.retrieve(object_to_delete)

    tombstone = Object(type="Tombstone", id=object_to_delete).build()

    if to_update_from_db:
        await actor.update(tombstone)

    data = copy_to_and_cc(to_update_from_db, item.data)

    await actor.update(data)
    item.data = data

    return item



import asyncio
import logging
import traceback

from bovine.activitystreams.utils import recipients_for_object, remove_public
from bovine.jsonld import with_external_context

logger = logging.getLogger(__name__)


async def get_inbox_for_recipient(actor, recipient):
    try:
        logger.info("Getting inbox for %s", recipient)
        remote = await actor.retrieve(recipient)
        return remote.get("inbox")
    except Exception as ex:
        logger.warning("Failed to fetch inbox for %s with %s", recipient, ex)
        return


async def send_to_inbox(actor, inbox, data):
    try:
        response = await actor.post(inbox, data)
        logger.info(await response.text())
    except Exception as ex:
        logger.warning("Sending to %s failed with %s", inbox, ex)
        for log_line in traceback.format_exc().splitlines():
            logger.warning(log_line)


async def determine_recipients(item, actor):
    recipients = recipients_for_object(item.data)
    recipients = remove_public(recipients)

    recipients = recipients | set(item.meta.get("additional_recipients", []))

    logger.debug("Recipients '%s'", "', '".join(recipients))

    endpoints = [x for x in recipients if x in actor.endpoints]

    if len(endpoints) > 0:
        recipients = {x for x in recipients if x not in endpoints}.union(
            await actor.resolve_endpoints(endpoints)
        )

    return {x for x in recipients if x != actor.actor_id}


async def send_outbox_item(item, actor):
    """Sends the item to all the specfied outboxes

    Collections are resolved if they are part of actor.endpoints and then
    resolved via actor.resolve_endpoints.

    inboxes from actors are fetched from the database or refetched
    """
    logger.info("Sending outbox item")

    recipients = await determine_recipients(item, actor)

    inboxes = [await get_inbox_for_recipient(actor, x) for x in recipients]
    inboxes = [x for x in inboxes if x]

    logger.info("Inboxes %s", " - ".join(inboxes))
    item.data = with_external_context(item.data)

    await asyncio.gather(*[send_to_inbox(actor, inbox, item.data) for inbox in inboxes])



from bovine_store import BovineStore
from contextlib import asynccontextmanager


@asynccontextmanager
async def bovine_actor(body, session):
    store = BovineStore(session=session)

    actor = await store.actor_for_name(body["bovine_name"])

    try:
        yield actor
    finally:
        ...



import logging

from bovine.activitystreams.utils import (
    actor_for_object,
    recipients_for_object,
    remove_public,
    is_public,
)

logger = logging.getLogger(__name__)


async def store_incoming(item, actor):
    owner = actor_for_object(item.data)
    public = is_public(item.data)
    if not public:
        recipients = remove_public(recipients_for_object(item.data))
        recipients.add(actor.actor_id)
        await actor.store_for(owner, item.data, visible_to=recipients)
        logger.info("Owner %s Recipients %s", owner, " | ".join(list(recipients)))
    else:
        await actor.store_for(owner, item.data, as_public=True)
        logger.info("Owner %s, public", owner)
    return item


async def add_incoming_to_inbox(item, actor):
    object_id = item.object_id()

    if object_id is None:
        logger.warning(
            "Tried to store object without id to inbox of %s", actor.actor_id
        )
        return item.data

    stored_item = await actor.add_to_inbox(object_id)

    item.meta["database_id"] = stored_item.id

    return item



import asyncio
import logging
import random

from bovine_store.actor.bovine_store_actor import BovineStoreActor

from .incoming import default_inbox_process
from .outgoing import default_async_outbox_process, default_outbox_process
from .send_item import send_outbox_item
from .types import ProcessingItem

logger = logging.getLogger(__name__)


async def process_inbox_item(
    item: ProcessingItem, actor: BovineStoreActor, retry=False
):
    """Asynchronous processing of an item arriving at the inbox"""

    try:
        logger.debug("Processing inbox request")
        result = await default_inbox_process(item, actor)
        return result
    except Exception:
        logger.error(">>>>> SOMETHING WENT WRONG IN PROCESSING <<<<<<")

        if item:
            item.dump()

        if not retry:
            wait_time = random.randint(3, 33)
            logger.error("Retrying in %d seconds", wait_time)
            await asyncio.sleep(wait_time)
            await process_inbox_item(item, actor, retry=True)


async def handle_outbox_item(
    item: ProcessingItem, actor: BovineStoreActor
) -> ProcessingItem | None:
    """Synchronous part of processing an outbox item. In particular
    it contains the logic to assign ids to the item"""

    return await default_outbox_process(item, actor)


async def process_outbox_item(item: ProcessingItem, actor: BovineStoreActor):
    """Asynchronous part of processing an outbox item. In particular
    it contains the logic of sending the message to all followers"""

    logger.debug("Processing outbox request")
    await default_async_outbox_process(item, actor)
    await send_outbox_item(item, actor)



import json
import logging
import uuid
from dataclasses import dataclass, field
from urllib.parse import urlparse

logger = logging.getLogger(__name__)


@dataclass
class ProcessingItem:
    """Represents an entity being processed"""

    submitter: str
    data: dict
    meta: dict = field(default_factory=dict)

    @property
    def submitter_domain(self):
        return urlparse(self.submitter).netloc

    def object_id(self):
        object_id = self.data.get("id")

        if object_id is None:
            object_id = f"remote://{str(uuid.uuid4())}"
            self.data["id"] = object_id

        return object_id

    def dump(self):
        logger.error(">>>> ITEM from %s <<<<", self.submitter)
        logger.error("data: %s", json.dumps(self.data))
        logger.error("meta: %s", json.dumps(self.meta))
        logger.error("   ")



import logging

from bovine.activitystreams.utils import id_for_object
from bovine.parse import Activity
from bovine_process.types import ProcessingItem

logger = logging.getLogger(__name__)


async def like_handler(item: ProcessingItem, actor) -> ProcessingItem:
    """Adds object to the likes collection of the liked object, if

    * object being liked is owner by the receiving actor"""

    object_to_like = item.data.get("object")
    obj = await actor.retrieve_own_object(id_for_object(object_to_like))
    if obj:
        obj_id = id_for_object(obj)
        logger.info("Like Handler %s", obj_id)
        await actor.add_to_interaction("likes", obj_id, item.data.get("id"))

    return item


async def announce_handler(item: ProcessingItem, actor) -> ProcessingItem:
    """Adds object to the shares collection of the announced object, if

    * object being announced is owner by the receiving actor"""

    object_to_share = item.data.get("object")
    obj = await actor.retrieve_own_object(id_for_object(object_to_share))
    if obj:
        obj_id = id_for_object(obj)
        logger.info("Announce Handler %s", obj_id)
        await actor.add_to_interaction("shares", obj_id, item.data.get("id"))

    return item


async def reply_handler(item: ProcessingItem, actor) -> ProcessingItem:
    """Adds object to the replies collection. Object being replied to
    is determined from `inReplyTo`. Reply is added if the object
    belongs to the receiving actor."""
    create = Activity(item.data, domain=item.submitter_domain)
    remote = await create.object_for_create(actor.retrieve)

    if not remote:
        return item

    if not remote.in_reply_to:
        return item

    obj = await actor.retrieve_own_object(remote.in_reply_to)

    if obj:
        obj_id = id_for_object(obj)
        logger.info("Reply Handler %s", obj_id)
        await actor.add_to_interaction("replies", obj_id, remote.identifier)

    return item


async def delete_reply_handler(item: ProcessingItem, actor) -> ProcessingItem:
    """If a reply is deleted, removes it from the replies collection"""

    remote = item.data.get("object")
    if not remote:
        return item

    await actor.remove_references(remote)

    return item


async def undo_handler(item: ProcessingItem, actor) -> ProcessingItem:
    """For an Undo of a Like, Dislike, Announce , they are removed from
    the appropriate collection."""

    object_to_undo = id_for_object(item.data.get("object"))
    obj = await actor.retrieve(object_to_undo)
    if obj is None:
        return item
    remote_actor = id_for_object(item.data.get("actor"))
    if obj.get("actor") != remote_actor:
        logger.error("Mismatching actor in undo from %s", remote_actor)
        return

    obj_type = obj.get("type")
    if obj_type in ["Like", "Dislike", "http://litepub.social/ns#EmojiReact"]:
        logger.info("Undo Handler for Like of %s", obj.get("id"))
        await actor.remove_from_interaction("likes", obj.get("object"), obj.get("id"))
    elif obj_type == "Announce":
        logger.info("Undo Handler for Like of %s", obj.get("id"))
        await actor.remove_from_interaction("shares", obj.get("object"), obj.get("id"))

    return item



import logging

from bovine.activitystreams.utils import is_public, recipients_for_object, remove_public

logger = logging.getLogger(__name__)


async def store_outgoing(item, actor):
    if is_public(item.data):
        await actor.store(item.data, as_public=True)
        logger.info("Owner %s, public", actor.actor_id)

    else:
        recipients = remove_public(recipients_for_object(item.data))
        recipients.add(actor.actor_id)

        await actor.store(item.data, visible_to=recipients)
        logger.info(
            "Owner %s Recipients %s", actor.actor_id, " | ".join(list(recipients))
        )

    return item


async def add_outgoing_to_outbox(item, actor):
    object_id = item.data.get("id")

    if object_id is None:
        logger.warning("Tried to store object with id to %s's outbox", actor.actor_id)
        return item

    stored_item = await actor.add_to_outbox(object_id)

    item.meta["database_id"] = stored_item.id

    return item



from bovine_store.utils.test import (  # noqa F401
    store,
    bovine_store_actor,
    bovine_admin_store,
)


from bovine_process.types import ProcessingItem

from .store_incoming import store_incoming


async def test_store_incoming(store, bovine_store_actor):  # noqa F801
    first_id = "https://my_domain/first"
    second_id = "https://my_domain/second"
    item = {
        "@context": "https://www.w3.org/ns/activitystreams",
        "id": first_id,
        "type": "Create",
        "object": {
            "type": "Note",
            "id": second_id,
        },
    }

    actor_id = bovine_store_actor.actor_object.id

    processing_item = ProcessingItem("tag:actor", item)

    result = await store_incoming(processing_item, bovine_store_actor)

    assert result == processing_item

    first = await store.retrieve(actor_id, first_id)
    second = await store.retrieve(actor_id, second_id)

    assert first["id"] == first_id
    assert first["object"] == second_id
    assert second["id"] == second_id



import logging

from urllib.parse import urlparse
from bovine.parse import Activity
from bovine_process.types import ProcessingItem

logger = logging.getLogger(__name__)


async def accept_follow(item: ProcessingItem, actor) -> ProcessingItem:
    accept = Activity(item.data, domain=urlparse(item.submitter).netloc)
    follow = await accept.accept_for_follow(actor.retrieve)

    if follow and follow.object_id == actor.actor_id:
        await actor.add_to_followers(follow.actor_id)
        logger.info(
            "Added %s to followers %s", follow.actor_id, actor.actor_object.followers
        )

    return item


async def undo_follow(item: ProcessingItem, actor) -> ProcessingItem:
    if item.data["type"] != "Undo":
        return item

    obj = item.data["object"]
    if isinstance(obj, str):
        obj = await actor.retrieve(obj)

    if obj["type"] != "Follow":
        return item

    remote_actor = obj.get("object")

    await actor.remove_from_following(remote_actor)

    logger.info(
        "Removed %s from following %s", remote_actor, actor.actor_object.following
    )

    return item



from .types import ProcessingItem


def test_get_object_id():
    def id_for_body(data):
        item = ProcessingItem("tag:author", data)
        return item.object_id()

    assert id_for_body({}).startswith("remote://")
    assert id_for_body({"id": "abc"}) == "abc"

    item = ProcessingItem("tag:author", {})

    assert item.object_id() == item.object_id()



import logging
import traceback

from .by_activity_type import ByActivityType

logger = logging.getLogger(__name__)


class ProcessorList:
    def __init__(self):
        self.processors = []

    def add(self, processor):
        self.processors.append(processor)
        return self

    def add_for_types(self, **kwargs):
        return self.add(ByActivityType(kwargs).act)

    async def apply(self, item, *arguments):
        working = item

        try:
            for processor in self.processors:
                working = await processor(working, *arguments)
                if not working:
                    return

            return working
        except Exception as ex:
            logger.warning(ex)
            for log_line in traceback.format_exc().splitlines():
                logger.error(log_line)
            raise ex



from unittest.mock import AsyncMock

from .processor_list import ProcessorList


async def test_processor_list():
    processor_list = ProcessorList()

    mock1 = AsyncMock()
    mock2 = AsyncMock()
    mock1.return_value = None

    processor_list.add(mock1).add(mock2)

    item = "item"

    result = await processor_list.apply(item)

    mock1.assert_awaited_once()
    mock2.assert_not_awaited()

    assert result is None
